from typing import Optional

from tablate.library.initializers.mappers.element.base.base_column_mapper import base_column_mapper
from tablate.library.initializers.mappers.element.base.base_frame_mapper import base_frame_mapper
from tablate.library.initializers.mappers.element.base.base_text_mapper import base_text_mapper
from tablate.library.initializers.mappers.element.html.html_column_mapper import html_column_mapper
from tablate.library.initializers.mappers.element.html.html_frame_mapper import html_frame_mapper
from tablate.library.initializers.mappers.element.html.html_text_mapper import html_text_mapper
from tablate.type.defaults import html_px_multiplier_default
from tablate.type.primitives import HtmlPxMultiplier
from tablate.type.type_base import FrameBase, TextBase, ColumnBase
from tablate.type.type_global import Options
from tablate.type.type_input import BaseStylesInput, HtmlStylesInput
from tablate.type.type_store import BaseFrameStore


def style_mapper(base_input: Optional[BaseStylesInput],
                 html_input: Optional[HtmlStylesInput],
                 frame_default: Optional[FrameBase],
                 column_default: Optional[ColumnBase],
                 text_default: Optional[TextBase],
                 html_px_multiplier: HtmlPxMultiplier,
                 options: Options) -> BaseFrameStore:

    if html_px_multiplier is None:
        html_px_multiplier = html_px_multiplier_default

    frame_styles = base_frame_mapper(frame_input=options.console.frame_styles)
    column_styles = base_column_mapper(columns_input=options.console.column_styles)
    text_styles = base_text_mapper(text_input=options.console.text_styles)

    if frame_default:
        frame_styles = base_frame_mapper(frame_input=frame_default, frame_defaults=frame_styles)
    if column_default:
        column_styles = base_column_mapper(columns_input=column_default, column_defaults=column_styles)
    if text_default:
        text_styles = base_text_mapper(text_input=text_default, text_defaults=text_styles)

    if base_input:
        if base_input.frame_styles:
            frame_styles = base_frame_mapper(frame_input=base_input.frame_styles, frame_defaults=frame_styles)
        if base_input.column_styles:
            column_styles = base_column_mapper(columns_input=base_input.column_styles, column_defaults=column_styles)
        if base_input.text_styles:
            text_styles = base_text_mapper(text_input=base_input.text_styles, text_defaults=text_styles)

    html_frame_styles = html_frame_mapper(base_frame_defaults=frame_styles, html_px_multiplier=html_px_multiplier)
    html_column_styles = html_column_mapper(base_column_defaults=column_styles)
    html_text_styles = html_text_mapper(base_text_defaults=text_styles)


    if html_input:
        if html_input.html_frame_styles:
            html_frame_styles = html_frame_mapper(html_frame_input=html_input.html_frame_styles,
                                                  base_frame_defaults=frame_styles)
        if html_input.html_column_styles:
            html_column_styles = html_column_mapper(html_columns_input=html_input.html_column_styles,
                                                    base_column_defaults=column_styles)
        if html_input.html_text_styles:
            html_text_styles = html_text_mapper(html_text_input=html_input.html_text_styles,
                                                base_text_defaults=text_styles,
                                                html_px_multiplier=html_px_multiplier)

    return BaseFrameStore(frame_styles=frame_styles,
                          column_styles=column_styles,
                          text_styles=text_styles,
                          html_frame_styles=html_frame_styles,
                          html_column_styles=html_column_styles,
                          html_text_styles=html_text_styles)



import time
from typing import List, Union

from tablate.classes.bases.TablateApiBase import TablateApi
from tablate.classes.bases.TablateApiSet import TablateSet
from tablate.library.initializers.grid_init import grid_init
from tablate.library.initializers.table_init import table_init
from tablate.library.initializers.text_init import text_init
from tablate.type.primitives import OuterBorder, FrameDivider, OuterPadding, OuterWidth, Multiline, TextColor, \
    TextAlign, TextStyle, MaxLines, HideHeader, Background, ColumnDivider, ColumnPadding, HtmlPxMultiplier, \
    BackgroundPadding, FrameName
from tablate.type.type_base import HtmlOuterBase, HtmlTableRowsBase, TableRowsBase, TextBase
from tablate.type.type_input import GridColumnInput, TableHeaderFrameStylesInput, \
    TableBodyFrameStylesInput, HtmlTableBodyFrameStylesInput, HtmlTableHeaderFrameStylesInput, \
    HtmlTableFrameStylesInput, TableColumnInput, \
    HtmlTextStylesInput, HtmlColumnStylesInput, ColumnStylesInput, TextStylesInput, TableRowsDataInputDict, \
    HtmlFrameStylesInput, FrameStylesInput, HtmlOuterStylesInput, HtmlTextFrameStylesInput, \
    HtmlGridFrameStylesInput


class Tablate(TablateSet):

    def __init__(self,
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,

                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_outer_styles: HtmlOuterStylesInput = None,

                 column_styles: ColumnStylesInput = None,
                 text_styles: TextStylesInput = None,

                 html_frame_styles: HtmlFrameStylesInput = None,
                 html_column_styles: HtmlColumnStylesInput = None,
                 html_text_styles: HtmlTextStylesInput = None) -> None:

        super().__init__(outer_border=outer_border,
                         outer_padding=outer_padding,
                         outer_width=outer_width,
                         frame_divider=frame_divider,
                         background=background,
                         background_padding=background_padding,
                         html_px_multiplier=html_px_multiplier,
                         html_styles=html_outer_styles,
                         column_styles=column_styles,
                         text_styles=text_styles,
                         html_frame_styles=html_frame_styles,
                         html_column_styles=html_column_styles,
                         html_text_styles=html_text_styles)

    def add_text_frame(self,
                       text: Union[str, int, float],

                       name: FrameName = None,

                       text_style: TextStyle = None,
                       text_align: TextAlign = None,
                       text_color: TextColor = None,

                       frame_divider: FrameDivider = None,
                       frame_padding: ColumnPadding = None,
                       background: Background = None,
                       background_padding: BackgroundPadding = None,
                       multiline: Multiline = None,
                       max_lines: MaxLines = None,

                       html_px_multiplier: HtmlPxMultiplier = None,
                       html_styles: HtmlTextFrameStylesInput = None) -> None:

        text_frame_store = text_init(text=text,
                                     name=name,
                                     frame_divider=frame_divider,
                                     frame_padding=frame_padding,
                                     background=background,
                                     background_padding=background_padding,
                                     multiline=multiline,
                                     max_lines=max_lines,
                                     text_style=text_style,
                                     text_align=text_align,
                                     text_color=text_color,
                                     html_px_multiplier=html_px_multiplier,
                                     html_styles=html_styles,
                                     options=self._options)

        self._frame_list.append(text_frame_store)

    def add_grid_frame(self,
                       columns: List[Union[str, GridColumnInput]],

                       name: FrameName = None,

                       frame_divider: FrameDivider = None,
                       background: Background = None,
                       background_padding: BackgroundPadding = None,
                       multiline: Multiline = None,
                       max_lines: MaxLines = None,

                       column_divider: ColumnDivider = None,
                       column_padding: ColumnPadding = None,

                       text_style: TextStyle = None,
                       text_align: TextAlign = None,
                       text_color: TextColor = None,

                       html_px_multiplier: HtmlPxMultiplier = None,
                       html_styles: HtmlGridFrameStylesInput = None) -> None:

        grid_frame_store = grid_init(columns=columns,
                                     name=name,
                                     frame_divider=frame_divider,
                                     background=background,
                                     background_padding=background_padding,
                                     multiline=multiline,
                                     max_lines=max_lines,
                                     column_divider=column_divider,
                                     column_padding=column_padding,
                                     text_style=text_style,
                                     text_align=text_align,
                                     text_color=text_color,
                                     html_styles=html_styles,
                                     html_px_multiplier=html_px_multiplier,
                                     options=self._options)

        self._frame_list.append(grid_frame_store)

    def add_table_frame(self,
                        columns: List[TableColumnInput],
                        rows: List[TableRowsDataInputDict],

                        name: FrameName = None,

                        frame_divider: FrameDivider = None,
                        multiline: Multiline = None,
                        max_lines: MaxLines = None,
                        background: Background = None,
                        background_padding: BackgroundPadding = None,

                        multiline_header: Multiline = None,
                        max_lines_header: MaxLines = None,
                        hide_header: HideHeader = None,

                        column_divider: ColumnDivider = None,
                        column_padding: ColumnPadding = None,
                        header_base_divider: FrameDivider = None,

                        row_line_divider: FrameDivider = None,
                        odd_row_background: Background = None,
                        even_row_background: Background = None,

                        text_style: TextStyle = None,
                        text_align: TextAlign = None,
                        text_color: TextColor = None,

                        header_styles: TableHeaderFrameStylesInput = None,
                        body_styles: TableBodyFrameStylesInput = None,

                        html_px_multiplier: HtmlPxMultiplier = None,
                        html_styles: HtmlTableFrameStylesInput = None,

                        html_header_styles: HtmlTableHeaderFrameStylesInput = None,
                        html_body_styles: HtmlTableBodyFrameStylesInput = None) -> None:

        table_header_store, table_body_store = table_init(columns=columns,
                                                          rows=rows,
                                                          name=name,
                                                          frame_divider=frame_divider,
                                                          multiline=multiline,
                                                          max_lines=max_lines,
                                                          background=background,
                                                          background_padding=background_padding,
                                                          multiline_header=multiline_header,
                                                          max_lines_header=max_lines_header,
                                                          hide_header=hide_header,
                                                          column_divider=column_divider,
                                                          column_padding=column_padding,
                                                          header_base_divider=header_base_divider,
                                                          row_line_divider=row_line_divider,
                                                          odd_row_background=odd_row_background,
                                                          even_row_background=even_row_background,
                                                          text_style=text_style,
                                                          text_align=text_align,
                                                          text_color=text_color,
                                                          header_styles=header_styles,
                                                          body_styles=body_styles,
                                                          html_px_multiplier=html_px_multiplier,
                                                          html_styles=html_styles,
                                                          html_header_styles=html_header_styles,
                                                          html_body_styles=html_body_styles,
                                                          options=self._options)

        if table_header_store:
            self._frame_list.append(table_header_store)

        self._frame_list.append(table_body_store)


########################################################################################################################
########################################################################################################################
########################################################################################################################

#
# column_list: List[Union[GridColumnInput, TableColumnInput]] = [
#     {
#         "string": "Column One",
#         "key": "Key One",
#         "width": "50%",
#         "divider": "thick",
#         "background": "blue"
#     },
#     {
#         "string": "Column Two",
#         "key": "Key Two",
#         "text_align": "right",
#         "divider": "double"
#
#     },
#     {
#         "string": "Column Three",
#         "key": "Key Three",
#         "text_align": "right"
#     }
# ]
#
# column_list2: List[Union[GridColumnInput, TableColumnInput]] = [
#     {
#         "string": "Column One",
#         "key": "Key One",
#         "width": "25%",
#         "divider": "thick",
#         "background": "blue"
#     },
#     {
#         "string": "Column Two",
#         "key": "Key Two",
#         "text_align": "right",
#         "divider": "double"
#
#     },
#     {
#         "string": "Column Three",
#         "key": "Key Three",
#         "text_align": "right"
#     }
# ]
#
# row_list = [
#     {
#         "Key One": "Some string value",
#         "Key Two": 4,
#         "Key Three": 9.6,
#         "divider": "thick"
#     },
#     {
#         "Key One": "Some other string value",
#         "Key Two": 5,
#         "Key Three": 4.6
#
#     },
#     {
#         "Key One": "Some final string value",
#         "Key Two": 100,
#         "Key Three": 328.832
#
#     }
# ]
# ########################################################################################################################
# html_global_styles_input = HtmlOuterBase(html_outer_border_style="thick",
#                                          html_outer_padding=16,
#                                          html_outer_width="100%")
# ########################################################################################################################
# frame_styles_input = FrameStylesInput(frame_divider="thick",
#                                       max_lines=None,
#                                       multiline=True,
#                                       background="yellow",
#                                       trunc_value="...")
# column_styles_input = ColumnStylesInput(divider="thick",
#                                         padding=1)
# text_styles_input = TextStylesInput(text_style="bold",
#                                     text_align="left",
#                                     text_color="green")
# ########################################################################################################################
# html_frame_styles_input = HtmlFrameStylesInput(html_frame_divider_style="thick",
#                                                html_max_lines=5,
#                                                html_multiline=True,
#                                                html_background="green")
# html_column_styles_input = HtmlColumnStylesInput(html_divider_style="thin",
#                                                  html_padding=6)
# html_text_styles_input = HtmlTextStylesInput(html_text_style="normal",
#                                              html_text_align="left",
#                                              html_text_color="red",
#                                              html_text_size=16)
# ########################################################################################################################
# table_rows_frame_styles_input = TableBodyFrameStylesInput(frame_styles=frame_styles_input,
#                                                           column_styles=column_styles_input,
#                                                           text_styles=text_styles_input,
#                                                           row_styles=TableRowsBase(row_line_divider="thin",
#                                                                                    odds_background="grey",
#                                                                                    evens_background="green"))
# header_frame_styles_input = TableHeaderFrameStylesInput(frame_styles=frame_styles_input,
#                                                         column_styles=column_styles_input,
#                                                         text_styles=text_styles_input)
# html_header_styles_input = HtmlTableHeaderFrameStylesInput(html_frame_styles=html_frame_styles_input,
#                                                            html_column_styles=html_column_styles_input,
#                                                            html_text_styles=html_text_styles_input)
# html_rows_styles_input = HtmlTableBodyFrameStylesInput(html_frame_styles=html_frame_styles_input,
#                                                        html_column_styles=html_column_styles_input,
#                                                        html_text_styles=html_text_styles_input,
#                                                        html_row_styles=HtmlTableRowsBase(
#                                                            html_row_line_divider_style="none"
#                                                        ))
# ########################################################################################################################
#
# start = time.perf_counter_ns()
# test = Tablate()
# # test = Tablate(outer_border="thin",
# #                      outer_padding=6,
# #                      outer_width=120,
# #                      html_outer_styles=html_global_styles_input,
# #                      frame_divider="double",
# #                      background="blue",
# #                      column_styles=column_styles_input,
# #                      text_styles=text_styles_input,
# #                      html_frame_styles=html_frame_styles_input,
# #                      html_column_styles=html_column_styles_input,
# #                      html_text_styles=html_text_styles_input)
#
# test.add_text_frame(text="Some String...",
#                     text_style="bold",
#                     text_align="left",
#                     text_color="red",
#                     frame_divider="thick",
#                     background="blue",
#                     multiline=True,
#                     max_lines=5)
#
# test.add_grid_frame(columns=column_list,
#                     frame_divider="thick",
#                     text_color="white",
#                     text_style="bold_underlined",
#                     background="dark_red",
#                     multiline=True,
#                     max_lines=5)
#
# test.add_table_frame(columns=column_list,
#                      rows=row_list,
#                      odd_row_background="red",
#                      text_color="yellow")
#
# test.add_grid_frame(columns=column_list)
#
# test.add_table_frame(columns=column_list,
#                      rows=row_list,
#                      name="Some Table",
#                      frame_divider="thick",
#                      background="blue",
#                      multiline=True,
#                      max_lines=5,
#                      multiline_header=False,
#                      max_lines_header=None,
#                      hide_header=False,
#
#                      column_divider="thin",
#                      column_padding=1,
#                      header_base_divider="thick",
#                      row_line_divider="thin",
#                      even_row_background="magenta",
#                      odd_row_background="blue",
#
#                      text_style="bold",
#                      text_align="center",
#                      text_color="blue",
#
#                      header_styles=header_frame_styles_input,
#                      body_styles=table_rows_frame_styles_input,
#
#                      html_styles=html_rows_styles_input,
#
#                      html_header_styles=html_header_styles_input,
#                      html_body_styles=html_rows_styles_input)
#
#
# test.add_text_frame(text="Some String...")
#
# test.add_grid_frame(columns=column_list)
#
# test.add_table_frame(columns=column_list2,
#                      rows=row_list,
#                      hide_header=True)
#
# test.add_grid_frame(columns=column_list, html_px_multiplier=3, text_color="blue")
#
# test.add_table_frame(columns=column_list,
#                      rows=row_list,
#                      text_color="grey")
#
#
# string = test.to_string()
# html = test.to_html()
# print(string)
# print(html)
#
# test.list_frames()
#
# test.get_frame(9).print()

column_list = [
    {
        "string": "Column One",
        "key": "one",
        "width": "50%",
        "divider": "thick",
    },
    {
        "string": "Column Two",
        "key": "two",
        "text_align": "right",

    },
    {
        "string": "Column Three",
        "key": "three",
        "text_align": "right"
    }
]

row_list = [
    {
        "one": "Some string value",
        "two": 4,
        "three": 9.6,
    },
    {
        "one": "Some other string value",
        "two": 5,
        "three": 4.6

    },
    {
        "one": "Some final string value",
        "two": 100,
        "three": 328.832

    }
]

tab = Tablate()

tab.add_table_frame(columns=column_list, rows=row_list, background="red")
print(tab.to_html())


from tablate.type.primitives import FrameNameList
from tablate.type.type_store import FrameStoreList
from tablate.type.type_global import Options


class TablateBase:

    _options: Options

    _frame_list: FrameStoreList

    _name_list: FrameNameList



from tablate.classes.options.html.style.subclasses.TextStyler import TextStyler


def html_text_formatter(text_styler: TextStyler, string: str) -> str:

    text_classnames = text_styler.generate_class_names()

    return f'<div><p class="{text_classnames}">{string}</p></div>'



from typing import List


def get_row_colspans(frame_columns: list, baseline_column_widths: List[int]) -> List[int]:
    return_row_spans = []
    column_accumulator = 0
    previous_column_index = -1
    for row_column in frame_columns:
        current_column_index = baseline_column_widths.index(row_column["baseline_width_percent"])
        column_accumulator += row_column["baseline_width_percent"]
        colspan = current_column_index - previous_column_index
        return_row_spans.append(colspan)
        previous_column_index = current_column_index
    return return_row_spans


from __future__ import annotations
#
from importlib.metadata import version
#
from tablate.api.Tablate import Tablate
from tablate.api.modules.Grid import Grid
from tablate.api.modules.Table import Table
from tablate.api.modules.Text import Text

# from tablate.api.functions import concat

try:
    __version__ = version('tablate')
except ModuleNotFoundError:
    pass



from tablate.classes.options.html.style.subclasses.ElementStyler import ElementStyler
from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.formatters.html.style.attributes.color import background_color_attr
from tablate.library.formatters.html.style.attributes.divider import divider_attr
from tablate.type.type_base import HtmlTableRowsBase


def style_row(row_store: HtmlTableRowsBase, row_styler: ElementStyler):
    row_divider_style = set_attr_resolver(instance=row_store, key="html_row_line_divider_style", default=None)
    row_divider_weight = set_attr_resolver(instance=row_store, key="html_row_line_divider_weight", default=None)
    odds_background = set_attr_resolver(instance=row_store, key="html_odds_background", default=None)
    evens_background = set_attr_resolver(instance=row_store, key="html_evens_background", default=None)
    if row_divider_style is not None and row_divider_weight is not None:
        row_styler.add_style_attribute("border-bottom",
                                       divider_attr(divider_style=row_divider_style,
                                                    divider_weight=row_divider_weight),
                                       sub_selector=":not(:last-child)")
    if odds_background is not None:
        row_styler.add_style_attribute("background-color",
                                       background_color_attr(odds_background),
                                       sub_selector=":nth-child(odd)")
    if evens_background is not None:
        row_styler.add_style_attribute("background-color",
                                       background_color_attr(evens_background),
                                       sub_selector=":nth-child(even)")


from typing import List

from tablate.library.ascii.chars.line_v import v_line
from tablate.library.checkers.is_last_element import is_last_element
from tablate.library.formatters.console.cell_string import cell_string_single_line
from tablate.library.formatters.console.row_outer_border import row_outer_border
from tablate.type.type_store import FrameStoreUnion
from tablate.type.type_global import Options


def column_console_multiline(formatted_columns_array: List[List[str]], frame_dict: FrameStoreUnion, options: Options):
    line_count = 0

    for column_item in formatted_columns_array:
        line_count = len(column_item) if len(column_item) > line_count else line_count

    line_count = line_count if frame_dict.frame_styles.max_lines is None \
                               or frame_dict.frame_styles.max_lines > line_count else frame_dict.frame_styles.max_lines

    return_string = ""

    for grid_line_index in range(line_count):
        grid_line_string = ""
        for column_index, column_item in enumerate(frame_dict.column_list):
            if grid_line_index < len(formatted_columns_array[column_index]):
                grid_line_string += formatted_columns_array[column_index][grid_line_index]
            else:
                grid_line_string += cell_string_single_line(string=column_item["string"],
                                                            column_item=column_item,
                                                            trunc_value=frame_dict.frame_styles.trunc_value)
            if not is_last_element(column_index, frame_dict.column_list):
                grid_line_string += v_line[column_item["divider"]]
        return_string += row_outer_border(row_string=grid_line_string, options=options)

    return return_string



from typing import Callable, Union, List

from tablate.classes.options.html.style.mixins.AddStyleMixin import AddStyleMixin
from tablate.classes.options.html.style.mixins.ClassNameMixin import ClassNameMixin
from tablate.classes.options.html.style.utilities.style_types import TableSelectorDict, SelectorDictUnion, StyleItem


class TableStyler(ClassNameMixin, AddStyleMixin):

    def __init__(self,
                 selector: TableSelectorDict,
                 create_style: Callable[[SelectorDictUnion, StyleItem], None]) -> None:
        self._selector_dict = selector
        self._create_style = create_style



from tablate.classes.options.html.style.subclasses.ElementStyler import ElementStyler
from tablate.library.formatters.html.element.column import html_column_head_formatter, html_column_foot_formatter
from tablate.library.formatters.html.element.frame import html_frame_head_formatter, html_frame_foot_formatter
from tablate.library.formatters.html.element.row import html_row_head_formatter, html_row_foot_formatter
from tablate.library.formatters.html.element.text import html_text_formatter
from tablate.library.formatters.html.style.elements.style_column import style_column_dict
from tablate.library.formatters.html.style.elements.style_frame import style_frame
from tablate.library.formatters.html.style.elements.style_rows import style_row
from tablate.type.type_global import Options

from tablate.library.calcs.get_row_colspan import get_row_colspans
from tablate.type.type_store import TableBodyFrameStore


def render_html_table_body(table_body_frame_store: TableBodyFrameStore, options: Options, frame_styler: ElementStyler):

    column_baselines = options.html.column_baselines

    colspans = get_row_colspans(table_body_frame_store.column_list, column_baselines)

    style_frame(frame_store=table_body_frame_store, frame_styler=frame_styler)
    style_row(row_store=table_body_frame_store.html_row_styles, row_styler=frame_styler.row())

    return_html = html_frame_head_formatter(frame_styler=frame_styler)

    for column_index, column_item in enumerate(table_body_frame_store.column_list):
        style_column_dict(column_dict=column_item, column_styler=frame_styler.column(column_index=column_index), html_px_multiplier=table_body_frame_store.html_frame_styles.html_px_multiplier)

    for row_index, row_item in enumerate(table_body_frame_store.row_list):

        row_styler = frame_styler.row(row_index)

        return_html += html_row_head_formatter(row_styler=row_styler)

        for row_column_index, row_column_item in enumerate(table_body_frame_store.column_list):

            column_styler = row_styler.column(row_column_index)

            return_html += html_column_head_formatter(column_styler=column_styler,
                                                      column_index=row_column_index,
                                                      colspans=colspans)

            text_styler = column_styler.text

            return_html += html_text_formatter(text_styler=text_styler,
                                               string=row_item[row_column_item["key"]])

            return_html += html_column_foot_formatter()

        return_html += html_row_foot_formatter()
    return_html += html_frame_foot_formatter()

    return return_html






from tablate.classes.options.html.style.subclasses.ElementStyler import ElementStyler
from tablate.type.primitives import HtmlRowGroupElement


def html_frame_head_formatter(frame_styler: ElementStyler, frame_element: HtmlRowGroupElement = "tbody") -> str:

    frame_classes = frame_styler.generate_class_names()

    return f'<{frame_element} class="{frame_classes}">'


def html_frame_foot_formatter(frame_element: HtmlRowGroupElement = "tbody") -> str:
    return f'</{frame_element}>'



from tablate.library.ascii.chars.line_h import h_line
from tablate.library.ascii.chars.line_v import v_line
from tablate.library.ascii.chars.matrix_cross import cross_matrix
from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.formatters.console.ascii_styler import ascii_text_styler, ascii_terminator
from tablate.library.formatters.console.concat_string import concat_string
from tablate.type.type_store import FrameStoreList


def list_frames(frame_list: FrameStoreList):
    print()
    # top_index = f"{' '}{h_line['thin'] * ((len(frame_list) // 10) + 3)}{cross_matrix['blank']['thin']['thin']}"
    # top_name = f"{h_line['thin'] * 22}{cross_matrix['blank']['thin']['thin']}"
    # top_type = f"{h_line['thin'] * 8}{cross_matrix['blank']['thin']['thin']}"
    # top_cols = f"{h_line['thin'] * 7}{cross_matrix['blank']['thin']['thin']}"
    # top_rows = f"{h_line['thin'] * 7}{cross_matrix['blank']['thin']['thin']}"
    # top_options = f"{h_line['thin'] * 42}"
    # print(f"{top_index}{top_name}{top_type}{top_cols}{top_rows}{top_options}")
    divider = v_line["thin"]
    padding = " "
    header_string = f"{padding} {' ' * ((len(frame_list) // 10) + 1)}{padding}"
    frame_name, name_ws = concat_string(string="Name",
                                        width=20,
                                        padding=0,
                                        trunc_value="")
    frame_name = ascii_text_styler(string=frame_name, column_dict={"text_style": "bold"}) + ascii_terminator()
    header_string += f"{divider}{padding}{frame_name}{' ' * name_ws}{padding}"
    frame_type, type_ws = concat_string(string="Type",
                                        width=6,
                                        padding=0,
                                        trunc_value="")
    frame_type = ascii_text_styler(string=frame_type, column_dict={"text_style": "bold"}) + ascii_terminator()
    header_string += f"{divider}{padding}{frame_type}{' ' * type_ws}{padding}"
    frame_cols, cols_ws = concat_string(string="Cols.",
                                        width=5,
                                        padding=0,
                                        trunc_value="")
    frame_cols = ascii_text_styler(string=frame_cols, column_dict={"text_style": "bold"}) + ascii_terminator()
    header_string += f"{divider}{padding}{frame_cols}{' ' * cols_ws}{padding}"
    frame_rows, rows_ws = concat_string(string="Rows",
                                        width=5,
                                        padding=0,
                                        trunc_value="")
    frame_rows = ascii_text_styler(string=frame_rows, column_dict={"text_style": "bold"}) + ascii_terminator()
    header_string += f"{divider}{padding}{' ' * rows_ws}{frame_rows}{padding}"
    frame_options = ascii_text_styler(string="Options", column_dict={"text_style": "bold"}) + ascii_terminator()
    header_string += f"{divider}{padding}{frame_options}{padding}"
    print(header_string)
    divider_index = f"{padding}{h_line['thin'] * ((len(frame_list) // 10) + 3)}{cross_matrix['thin']['thin']['thin']}"
    divider_name = f"{h_line['thin'] * 22}{cross_matrix['thin']['thin']['thin']}"
    divider_type = f"{h_line['thin'] * 8}{cross_matrix['thin']['thin']['thin']}"
    divider_cols = f"{h_line['thin'] * 7}{cross_matrix['thin']['thin']['thin']}"
    divider_rows = f"{h_line['thin'] * 7}{cross_matrix['thin']['thin']['thin']}"
    divider_options = f"{h_line['thin'] * 42}"
    print(f"{divider_index}{divider_name}{divider_type}{divider_cols}{divider_rows}{divider_options}")
    frame_index = 0
    for frame_item in frame_list:
        if frame_item.type == "table_header":
            continue
        item_string = f"{padding}"
        index, index_ws = concat_string(string=frame_index,
                                        width=(len(frame_list) // 10) + 1,
                                        padding=0,
                                        trunc_value="")
        item_string += f"{padding}{' ' * index_ws}{index}{padding}"
        item_name, name_ws = concat_string(string=frame_item.name,
                                           width=20,
                                           padding=0,
                                           trunc_value="...")
        item_string += f"{divider}{padding}{item_name}{' ' * name_ws}{padding}"
        item_type = frame_item.type.capitalize() if frame_item.type != "table_body" else "Table"
        item_type, type_ws = concat_string(string=item_type,
                                           width=6,
                                           padding=0,
                                           trunc_value="")
        item_string += f"{divider}{padding}{item_type}{' ' * type_ws}{padding}"
        item_cols, cols_ws = concat_string(string=len(frame_item.column_list),
                                           width=5,
                                           padding=0,
                                           trunc_value="")
        item_string += f"{divider}{padding}{' ' * cols_ws}{item_cols}{padding}"
        item_rows_list = set_attr_resolver(instance=frame_item, key="row_list", default=[])
        item_rows, rows_ws = concat_string(string=len(item_rows_list),
                                           width=5,
                                           padding=0,
                                           trunc_value="")
        item_rows = item_rows if frame_item.type == "table_body" or frame_item.type == "table_header" else "-"
        item_string += f"{divider}{padding}{' ' * rows_ws}{item_rows}{padding}"
        item_options_list = []
        item_options_list.append("hide_header") if hasattr(frame_item, "hide_header") and frame_item.hide_header is True else None
        item_options_list.append("multiline") if frame_item.frame_styles.multiline is True else None
        item_options_list.append(
            f"max_lines: {frame_item.frame_styles.max_lines}") if frame_item.frame_styles.multiline is True else None
        item_string += f"{divider}{padding}{'; '.join(item_options_list)}{padding}"
        print(item_string)
        frame_index += 1
    print()


import time
from typing import List, Union

from tablate.classes.bases.TablateApiBase import TablateApi
from tablate.classes.bases.TablateApiSet import TablateSet
from tablate.library.initializers.grid_init import grid_init
from tablate.library.initializers.table_init import table_init
from tablate.library.initializers.text_init import text_init
from tablate.type.primitives import OuterBorder, FrameDivider, OuterPadding, OuterWidth, Multiline, TextColor, \
    TextAlign, TextStyle, MaxLines, HideHeader, Background, ColumnDivider, ColumnPadding, HtmlPxMultiplier, \
    BackgroundPadding, FrameName
from tablate.type.type_base import HtmlOuterBase, HtmlTableRowsBase, TableRowsBase, TextBase
from tablate.type.type_input import GridColumnInput, TableHeaderFrameStylesInput, \
    TableBodyFrameStylesInput, HtmlTableBodyFrameStylesInput, HtmlTableHeaderFrameStylesInput, \
    HtmlTableFrameStylesInput, TableColumnInput, \
    HtmlTextStylesInput, HtmlColumnStylesInput, ColumnStylesInput, TextStylesInput, TableRowsDataInputDict, \
    HtmlFrameStylesInput, FrameStylesInput, HtmlOuterStylesInput, HtmlTextFrameStylesInput, \
    HtmlGridFrameStylesInput


class Tablate(TablateSet):

    def __init__(self,
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,

                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_outer_styles: HtmlOuterStylesInput = None,

                 column_styles: ColumnStylesInput = None,
                 text_styles: TextStylesInput = None,

                 html_frame_styles: HtmlFrameStylesInput = None,
                 html_column_styles: HtmlColumnStylesInput = None,
                 html_text_styles: HtmlTextStylesInput = None) -> None:

        super().__init__(outer_border=outer_border,
                         outer_padding=outer_padding,
                         outer_width=outer_width,
                         frame_divider=frame_divider,
                         background=background,
                         background_padding=background_padding,
                         html_px_multiplier=html_px_multiplier,
                         html_styles=html_outer_styles,
                         column_styles=column_styles,
                         text_styles=text_styles,
                         html_frame_styles=html_frame_styles,
                         html_column_styles=html_column_styles,
                         html_text_styles=html_text_styles)

    def add_text_frame(self,
                       text: Union[str, int, float],

                       name: FrameName = None,

                       text_style: TextStyle = None,
                       text_align: TextAlign = None,
                       text_color: TextColor = None,

                       frame_divider: FrameDivider = None,
                       frame_padding: ColumnPadding = None,
                       background: Background = None,
                       background_padding: BackgroundPadding = None,
                       multiline: Multiline = None,
                       max_lines: MaxLines = None,

                       html_px_multiplier: HtmlPxMultiplier = None,
                       html_styles: HtmlTextFrameStylesInput = None) -> None:
        if name is not None and name not in self._name_list:
            self._name_list.append(name)
            text_frame_store = text_init(text=text,
                                         name=name,
                                         frame_divider=frame_divider,
                                         frame_padding=frame_padding,
                                         background=background,
                                         background_padding=background_padding,
                                         multiline=multiline,
                                         max_lines=max_lines,
                                         text_style=text_style,
                                         text_align=text_align,
                                         text_color=text_color,
                                         html_px_multiplier=html_px_multiplier,
                                         html_styles=html_styles,
                                         options=self._options)

            self._frame_list.append(text_frame_store)

    def add_grid_frame(self,
                       columns: List[Union[str, GridColumnInput]],

                       name: FrameName = None,

                       frame_divider: FrameDivider = None,
                       background: Background = None,
                       background_padding: BackgroundPadding = None,
                       multiline: Multiline = None,
                       max_lines: MaxLines = None,

                       column_divider: ColumnDivider = None,
                       column_padding: ColumnPadding = None,

                       text_style: TextStyle = None,
                       text_align: TextAlign = None,
                       text_color: TextColor = None,

                       html_px_multiplier: HtmlPxMultiplier = None,
                       html_styles: HtmlGridFrameStylesInput = None) -> None:
        if name is not None and name not in self._name_list:
            self._name_list.append(name)
            grid_frame_store = grid_init(columns=columns,
                                         name=name,
                                         frame_divider=frame_divider,
                                         background=background,
                                         background_padding=background_padding,
                                         multiline=multiline,
                                         max_lines=max_lines,
                                         column_divider=column_divider,
                                         column_padding=column_padding,
                                         text_style=text_style,
                                         text_align=text_align,
                                         text_color=text_color,
                                         html_styles=html_styles,
                                         html_px_multiplier=html_px_multiplier,
                                         options=self._options)

            self._frame_list.append(grid_frame_store)

    def add_table_frame(self,
                        columns: List[TableColumnInput],
                        rows: List[TableRowsDataInputDict],

                        name: FrameName = None,

                        frame_divider: FrameDivider = None,
                        multiline: Multiline = None,
                        max_lines: MaxLines = None,
                        background: Background = None,
                        background_padding: BackgroundPadding = None,

                        multiline_header: Multiline = None,
                        max_lines_header: MaxLines = None,
                        hide_header: HideHeader = None,

                        column_divider: ColumnDivider = None,
                        column_padding: ColumnPadding = None,
                        header_base_divider: FrameDivider = None,

                        row_line_divider: FrameDivider = None,
                        odd_row_background: Background = None,
                        even_row_background: Background = None,

                        text_style: TextStyle = None,
                        text_align: TextAlign = None,
                        text_color: TextColor = None,

                        header_styles: TableHeaderFrameStylesInput = None,
                        body_styles: TableBodyFrameStylesInput = None,

                        html_px_multiplier: HtmlPxMultiplier = None,
                        html_styles: HtmlTableFrameStylesInput = None,

                        html_header_styles: HtmlTableHeaderFrameStylesInput = None,
                        html_body_styles: HtmlTableBodyFrameStylesInput = None) -> None:

        if name is not None and name not in self._name_list:
            self._name_list.append(name)

            table_header_store, table_body_store = table_init(columns=columns,
                                                              rows=rows,
                                                              name=name,
                                                              frame_divider=frame_divider,
                                                              multiline=multiline,
                                                              max_lines=max_lines,
                                                              background=background,
                                                              background_padding=background_padding,
                                                              multiline_header=multiline_header,
                                                              max_lines_header=max_lines_header,
                                                              hide_header=hide_header,
                                                              column_divider=column_divider,
                                                              column_padding=column_padding,
                                                              header_base_divider=header_base_divider,
                                                              row_line_divider=row_line_divider,
                                                              odd_row_background=odd_row_background,
                                                              even_row_background=even_row_background,
                                                              text_style=text_style,
                                                              text_align=text_align,
                                                              text_color=text_color,
                                                              header_styles=header_styles,
                                                              body_styles=body_styles,
                                                              html_px_multiplier=html_px_multiplier,
                                                              html_styles=html_styles,
                                                              html_header_styles=html_header_styles,
                                                              html_body_styles=html_body_styles,
                                                              options=self._options)

            if table_header_store:
                self._frame_list.append(table_header_store)

            self._frame_list.append(table_body_store)




from typing import Optional

from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import background_default, row_line_divider_default, html_divider_weight_default
from tablate.type.type_base import HtmlTableRowsBase, TableRowsBase
from tablate.type.type_store import HtmlTableRowsStore


def html_rows_mapper(html_rows_input: Optional[HtmlTableRowsBase] = None,
                     html_rows_defaults: Optional[HtmlTableRowsBase] = None,
                     base_rows_defaults: Optional[TableRowsBase] = None) -> HtmlTableRowsStore:
    html_row_line_divider_weight = set_attr_resolver(
        instance=html_rows_input,
        key="html_row_line_divider_weight",
        default=set_attr_resolver(
            instance=html_rows_defaults,
            key="html_row_line_divider_weight",
            default=html_divider_weight_default))
    html_row_line_divider_style = set_attr_resolver(
        instance=html_rows_input,
        key="html_row_line_divider_style",
        default=set_attr_resolver(
            instance=html_rows_defaults,
            key="html_row_line_divider_style",
            default=set_attr_resolver(
                instance=base_rows_defaults,
                key="row_line_divider",
                default=row_line_divider_default)))
    html_odds_background = set_attr_resolver(
        instance=html_rows_input,
        key="html_odds_background",
        default=set_attr_resolver(
            instance=html_rows_defaults,
            key="html_odds_background",
            default=set_attr_resolver(
                instance=base_rows_defaults,
                key="odds_background",
                default=background_default)))
    html_evens_background = set_attr_resolver(
        instance=html_rows_input,
        key="html_evens_background",
        default=set_attr_resolver(
            instance=html_rows_defaults,
            key="html_evens_background",
            default=set_attr_resolver(
                instance=base_rows_defaults,
                key="evens_background",
                default=background_default)))

    rows_return = HtmlTableRowsStore(html_row_line_divider_weight=html_row_line_divider_weight,
                                     html_row_line_divider_style=html_row_line_divider_style,
                                     html_odds_background=html_odds_background,
                                     html_evens_background=html_evens_background)

    return rows_return



def text_align_attr(text_align: str):
    pass


def text_style_attr(text_style: str):
    pass


def text_color_attr(text_color: str):
    pass


def text_size_attr(text_size: str, px_multiplier: int = None):
    if text_size is None:
        text_size = html_text_size_default
    if px_multiplier is not None:
        text_size = text_size * px_multiplier
    return f"{text_size}px"



import math
from typing import List, overload, TypeVar, Type

from tablate.type.type_global import Options
from tablate.type.type_input import BaseColumnInput, GridColumnInput, TableColumnInput


def calc_column_width_error():
    raise Exception(
        f"Error computing width values. Please input check widths.")


@overload
def calc_column_widths(columns: List[BaseColumnInput], options: Options) -> List[BaseColumnInput]:
    ...


@overload
def calc_column_widths(columns: List[GridColumnInput], options: Options) -> List[GridColumnInput]:
    ...


@overload
def calc_column_widths(columns: List[TableColumnInput], options: Options) -> List[TableColumnInput]:
    ...


T = TypeVar("T", BaseColumnInput, GridColumnInput, TableColumnInput)


def calc_column_widths(columns: List[Type[T]], options: Options) -> List[T]:

    outer_width = options.console.outer_styles.outer_width

    total_defined_column_width = 0
    undefined_width_columns = []

    error = False

    for column_index, column_item in enumerate(columns):

        if "width" in column_item:
            validated_value = False
            if type(column_item["width"]) == str and column_item["width"][-1] == "%":
                try:
                    width_percentage = math.floor(outer_width / (100 / int(column_item["width"][0:-1]))) - 2
                    column_item["width"] = width_percentage
                    validated_value = True
                except TypeError:
                    error = True
            if type(column_item["width"]) == int:
                # if column_item["width"] < min_column_width:
                #     error = True
                total_defined_column_width += column_item["width"]
                validated_value = True
            if not validated_value:
                error = True
        else:
            undefined_width_columns.append(column_index)

    if error:
        calc_column_width_error()

    total_column_width = 0

    if len(undefined_width_columns):
        calculated_width_total = outer_width - (total_defined_column_width + len(columns) + 1)
        calculated_column_width = math.floor((calculated_width_total / len(undefined_width_columns)))
        # if calculated_column_width < min_column_width:
        #     error = True
        for column_index, column_item in enumerate(columns):
            if column_index in undefined_width_columns:
                column_item["width"] = calculated_column_width
            total_column_width += column_item["width"]
    else:
        total_column_width = total_defined_column_width

    if error:
        calc_column_width_error()
    if total_column_width + len(columns) + 1 > outer_width:
        calc_column_width_error()
    if total_column_width + len(columns) + 1 < outer_width:
        if len(undefined_width_columns) > 0:
            columns[undefined_width_columns[0]]["width"] += (outer_width - total_column_width) - (
                    len(columns) + 1)
        else:
            calc_column_width_error()

    return columns




from tablate.classes.options.html.style.subclasses.ElementStyler import ElementStyler
from tablate.library.calcs.get_row_colspan import get_row_colspans
from tablate.library.formatters.html.element.column import html_column_head_formatter, html_column_foot_formatter
from tablate.library.formatters.html.element.frame import html_frame_head_formatter, html_frame_foot_formatter
from tablate.library.formatters.html.element.row import html_row_head_formatter, html_row_foot_formatter
from tablate.library.formatters.html.element.text import html_text_formatter
from tablate.library.formatters.html.style.elements.style_column import style_column_dict
from tablate.library.formatters.html.style.elements.style_frame import style_frame
from tablate.type.primitives import HtmlRowGroupElement, HtmlCellElement, HtmlFrameType
from tablate.type.type_store import BaseFrameStore, FrameStoreUnion
from tablate.type.type_global import Options


def render_html_column(frame_dict: FrameStoreUnion, options: Options, frame_styler: ElementStyler, frame_type: HtmlFrameType = "body") -> str:

    frame_element: HtmlRowGroupElement
    column_element: HtmlCellElement

    if frame_type == "head":
        frame_element = "tbody"
        column_element = 'th'
    else:
        frame_element = "tbody"
        column_element = 'td'

    column_baselines = options.html.column_baselines
    html_px = options.html.html_outer_styles.html_px_multiplier

    colspans = get_row_colspans(frame_dict.column_list, column_baselines)

    style_frame(frame_store=frame_dict, frame_styler=frame_styler)

    return_html = html_frame_head_formatter(frame_styler=frame_styler, frame_element=frame_element)

    return_html += html_row_head_formatter(frame_styler.row(0))

    for column_index, column_item in enumerate(frame_dict.column_list):

        column_styler = frame_styler.column(column_index)

        style_column_dict(column_dict=column_item,
                          column_styler=column_styler,
                          html_px_multiplier=frame_dict.html_frame_styles.html_px_multiplier)

        return_html += html_column_head_formatter(column_styler=column_styler,
                                                  column_index=column_index,
                                                  colspans=colspans,
                                                  column_element=column_element)

        text_styler = column_styler.text

        return_html += html_text_formatter(text_styler=text_styler, string=column_item["string"])

        return_html += html_column_foot_formatter(column_element=column_element)
    return_html += html_row_foot_formatter()
    return_html += html_frame_foot_formatter(frame_element=frame_element)

    return return_html



from dataclasses import dataclass
from typing import Literal, Union, List, Dict

from tablate.type.primitives import TextString, TableRowKey, HtmlPxMultiplier, BackgroundPadding, FrameName, HideHeader
from tablate.type.type_base import ColumnBase, TextBase, HtmlTextBase, FrameBase, HtmlColumnBase, \
    HtmlTableRowsBase, TableRowsBase, HtmlFrameBase, HtmlOuterBase, OuterBase
from tablate.type.type_input import GridColumnInput, TableColumnInput, BaseColumnInput

########################################################################################################################
# FrameDicts ###########################################################################################################
########################################################################################################################


@dataclass
class OuterStore(OuterBase):
    background_padding: BackgroundPadding = None


@dataclass
class HtmlOuterStore(HtmlOuterBase):
    html_px_multiplier: HtmlPxMultiplier = None


@dataclass
class HtmlFrameStore(HtmlFrameBase):
    html_px_multiplier: HtmlPxMultiplier = None


FrameStore = FrameBase
ColumnStore = ColumnBase
TextStore = TextBase

HtmlColumnStore = HtmlColumnBase
HtmlTextStore = HtmlTextBase

TableRowsStore = TableRowsBase
HtmlTableRowsStore = HtmlTableRowsBase


class BaseColumnStore(BaseColumnInput):
    background_padding: BackgroundPadding


@dataclass
class BaseFrameStore:
    frame_styles: FrameStore
    column_styles: ColumnStore
    text_styles: TextStore
    html_frame_styles: HtmlFrameStore
    html_column_styles: HtmlColumnStore
    html_text_styles: HtmlTextStore


# Grid FrameDict #######################################################################################################


GridColumnStore = GridColumnInput


@dataclass()
class GridFrameStore(BaseFrameStore):
    type: Union[Literal["grid"], Literal["text"]]
    name: FrameName
    column_list: List[GridColumnStore]


# Table FrameDict ######################################################################################################

@dataclass()
class TableHeaderFrameStore(BaseFrameStore):
    type: Literal["table_header"]
    name: FrameName
    column_list: List[TableColumnInput]


@dataclass()
class TableBodyFrameStore(BaseFrameStore):
    type: Literal["table_body"]
    name: FrameName
    hide_header: HideHeader
    column_list: List[TableColumnInput]
    row_list: List[Dict[TableRowKey, TextString]]
    row_styles: TableRowsStore
    html_row_styles: HtmlTableRowsStore


# FrameDict List #######################################################################################################

FrameStoreUnion = Union[GridFrameStore, TableHeaderFrameStore, TableBodyFrameStore]
FrameStoreList = List[FrameStoreUnion]





from tablate.library.formatters.html.style.attributes.divider import divider_attr
from tablate.type.type_global import Options


def render_html_head(options: Options) -> str:

    outer_padding = options.html.html_outer_styles.html_outer_padding
    frame_border_style = options.html.html_outer_styles.html_outer_border_style
    frame_border_weight = options.html.html_outer_styles.html_outer_border_weight

    html_width = options.html.html_outer_styles.html_outer_width
    column_baselines = options.html.column_baselines

    margin_left_px = outer_padding
    margin_string = f'0 {margin_left_px}px'

    border_string = divider_attr(divider_style=frame_border_style, divider_weight=frame_border_weight)

    options.html.styler.add_global_style_attribute("font-family", "'Roboto', sans-serif")
    options.html.styler.add_global_style_attribute("box-sizing", "border-box")
    options.html.styler.add_global_style_attribute("margin", 0)
    options.html.styler.add_global_style_attribute("padding", 0)

    options.html.styler.wrapper.add_style_attribute("width", f"calc({html_width} - {margin_left_px * 2}px)")
    options.html.styler.wrapper.add_style_attribute("margin", margin_string)

    options.html.styler.table.add_style_attribute("width", "100%")
    options.html.styler.table.add_style_attribute("border", border_string)

    options.html.styler.table.add_style_attribute("border-collapse", "collapse")

    wrapper_classes = options.html.styler.wrapper.generate_class_names()
    table_classes = options.html.styler.table.generate_class_names()

    return_string = ''

    return_string += f'<div class="{wrapper_classes}">'
    return_string += f'<table class="{table_classes}">'

    return_string += f'<colgroup>'

    previous_baseline_value = 0
    for baseline_column_width in column_baselines:
        column_width = baseline_column_width - previous_baseline_value
        return_string += f'<col style="width:{column_width}%;">'
        previous_baseline_value = baseline_column_width
    return_string += f'</colgroup>'

    return return_string




from typing import List

from tablate.classes.helpers.list_frame import list_frames
from tablate.library.ascii.chars.line_h import h_line
from tablate.library.ascii.chars.line_v import v_line
from tablate.library.ascii.chars.matrix_cross import cross_matrix
from tablate.library.ascii.chars.matrix_side import left_side_matrix
from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.formatters.console.ascii_styler import ascii_text_styler, ascii_terminator
from tablate.library.formatters.console.concat_string import concat_string
from tablate.library.initializers.mappers.element.base.base_column_mapper import base_column_mapper
from tablate.library.initializers.mappers.element.base.base_frame_mapper import base_frame_mapper
from tablate.library.initializers.mappers.element.base.base_text_mapper import base_text_mapper
from tablate.library.initializers.mappers.element.html.html_column_mapper import html_column_mapper
from tablate.library.initializers.mappers.element.html.html_frame_mapper import html_frame_mapper
from tablate.library.initializers.mappers.element.html.html_outer_mapper import html_outer_mapper
from tablate.library.initializers.mappers.element.html.html_text_mapper import html_text_mapper
from tablate.type.type_base import FrameBase, TextBase, ColumnBase
from tablate.type.type_global import HtmlOptions

from tablate.classes.bases.TablateBase import TablateBase
from tablate.library.renderers.console.render_console import render_console
from tablate.library.renderers.html.render_html import render_html
from tablate.type.defaults import outer_border_default, outer_padding_default, outer_width_default, \
    html_px_multiplier_default, background_padding_default
from tablate.type.primitives import OuterBorder, FrameDivider, OuterWidth, OuterPadding, Background, HtmlPxMultiplier, \
    BackgroundPadding
from tablate.type.type_global import Options, ConsoleOptions
from tablate.type.type_input import HtmlOuterStylesInput, ColumnStylesInput, TextStylesInput, \
    HtmlFrameStylesInput, HtmlColumnStylesInput, HtmlTextStylesInput, BaseColumnInput, TableRowsDataInputDict
from tablate.type.type_store import OuterStore, TableHeaderFrameStore, TableBodyFrameStore


class TablateApi(TablateBase):

    def __init__(self,
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,

                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlOuterStylesInput = None,

                 column_styles: ColumnStylesInput = None,
                 text_styles: TextStylesInput = None,

                 html_frame_styles: HtmlFrameStylesInput = None,

                 html_column_styles: HtmlColumnStylesInput = None,
                 html_text_styles: HtmlTextStylesInput = None) -> None:

        self._name_list = []

        default_outer_styles = OuterStore(outer_border=outer_border if outer_border else outer_border_default,
                                          outer_padding=outer_padding if outer_padding else outer_padding_default,
                                          outer_width=outer_width if outer_width else outer_width_default,
                                          background_padding=background_padding if background_padding else background_padding_default)

        default_frame_styles = base_frame_mapper(frame_input=FrameBase(frame_divider=frame_divider,
                                                                       background=background),
                                                 frame_defaults=FrameBase())

        default_column_styles = base_column_mapper(columns_input=column_styles,
                                                   column_defaults=ColumnBase())

        default_text_styles = base_text_mapper(text_input=text_styles,
                                               text_defaults=TextBase())

        console_options = ConsoleOptions(outer_styles=default_outer_styles,
                                         frame_styles=default_frame_styles,
                                         column_styles=default_column_styles,
                                         text_styles=default_text_styles)

        if html_px_multiplier is None:
            html_px_multiplier = html_px_multiplier_default

        default_html_outer = html_outer_mapper(html_outer_input=html_styles,
                                               base_outer_defaults=default_outer_styles,
                                               html_px_multiplier=html_px_multiplier)
        default_html_frame = html_frame_mapper(html_frame_input=html_frame_styles,
                                               base_frame_defaults=default_frame_styles)
        default_html_column = html_column_mapper(html_columns_input=html_column_styles,
                                                 base_column_defaults=default_column_styles)
        default_html_text = html_text_mapper(html_text_input=html_text_styles,
                                             base_text_defaults=default_text_styles,
                                             html_px_multiplier=html_px_multiplier)

        html_options = HtmlOptions(html_outer_styles=default_html_outer,
                                   html_frame_styles=default_html_frame,
                                   html_column_styles=default_html_column,
                                   html_text_styles=default_html_text,
                                   css_injection="",
                                   column_baselines=[],
                                   styler=None)

        self._options: Options = Options(console=console_options, html=html_options)

        self._frame_list = []

    # def __is_ipython(self):
    #     return hasattr(__builtins__, "__IPYTHON__")

    # def list_frames(self):
    #     print_frame_list(self._frame_list)

    def to_string(self) -> str:
        return render_console(frame_list=self._frame_list, options=self._options)

    def print(self) -> None:
        print(self.to_string())

    def __repr__(self) -> str:
        return self.to_string()

    def to_html(self) -> str:
        return render_html(frame_list=self._frame_list, options=self._options)

    def _repr_html_(self) -> str:
        return self.to_html()



from typing import Union, Optional

from tablate.classes.bases.TablateApiBase import TablateApi
from tablate.library.initializers.text_init import text_init
from tablate.type.primitives import TextAlign, FrameDivider, OuterBorder, FrameName, TextStyle, TextColor, Background, \
    BackgroundPadding, Multiline, MaxLines, HtmlPxMultiplier, OuterPadding, OuterWidth, ColumnPadding
from tablate.type.type_input import HtmlTextFrameStylesInput, HtmlOuterStylesInput


class Text(TablateApi):

    def __init__(self,
                 # TablateText args
                 text: Union[str, int, float],

                 name: FrameName = None,

                 text_style: TextStyle = None,
                 text_align: TextAlign = None,
                 text_color: TextColor = None,
                 frame_padding: ColumnPadding = None,
                 frame_divider: FrameDivider = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,
                 multiline: Multiline = None,
                 max_lines: MaxLines = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlTextFrameStylesInput = None,
                 # TablateApi arge
                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,
                 html_outer_styles: HtmlOuterStylesInput = None) -> None:
        TablateApi.__init__(self=self,
                            outer_border=outer_border,
                            outer_padding=outer_padding,
                            frame_divider=frame_divider,
                            outer_width=outer_width,
                            html_styles=html_outer_styles)

        text_dict = text_init(text=text,
                              name=name,
                              text_style=text_style,
                              text_align=text_align,
                              text_color=text_color,
                              frame_divider=frame_divider,
                              frame_padding=frame_padding,
                              background=background,
                              background_padding=background_padding,
                              multiline=multiline,
                              max_lines=max_lines,
                              html_px_multiplier=html_px_multiplier,
                              html_styles=html_styles,
                              options=self._options)

        self._frame_list.append(text_dict)



from dataclasses import dataclass
from typing import Any, Type


def set_attr_resolver(instance: Any, key: str, default: Any):
    return instance.__getattribute__(key) if hasattr(instance, key) and instance.__getattribute__(key) is not None else default




import inspect
from typing import NamedTuple

class Point(NamedTuple):
    x: int
    y: int = 1  # Set default value

def f(a:str, b: int, c: bool = True):
    pass

func_sigs = inspect.getfullargspec(f)

print(func_sigs)

def some_func(arg: str = None):
    print(arg)





from tablate.library.ascii.colors.background import ascii_background_colors
from tablate.library.ascii.colors.characters import ascii_character_colors
from tablate.library.ascii.styles.styles import ascii_text_styles
from tablate.library.checkers.set_key_resolver import set_key_resolver
from tablate.type.type_input import BaseColumnStyles


def ascii_background_styler(string: str, column_dict: BaseColumnStyles):
    background = ascii_background_colors[column_dict["background"]]
    return f"\033[{background}m{string}"


def ascii_text_styler(string: str, column_dict: BaseColumnStyles):
    text_style = ascii_text_styles[set_key_resolver(instance=column_dict, key="text_style", default="normal")]
    text_color = ascii_character_colors[set_key_resolver(instance=column_dict, key="text_color", default="default")]
    if text_color != "" or text_style != "":
        return f"\033[{text_style}{';' if text_color != '' and text_style != '' else ''}{text_color}m{string}\033[24m"
    else:
        return string

def ascii_terminator():
    return "\033[0m"




from typing import Optional

from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import background_default, max_lines_default, \
    multiline_default, trunc_value_default, frame_divider_default, background_padding_default, table_multiline_default
from tablate.type.type_base import FrameBase
from tablate.type.type_store import FrameStore


def base_frame_mapper(frame_input: Optional[FrameBase] = None,
                      frame_defaults: Optional[FrameBase] = None) -> FrameStore:
    frame_return = FrameStore(frame_divider=set_attr_resolver(instance=frame_input,
                                                              key="frame_divider",
                                                              default=set_attr_resolver(instance=frame_defaults,
                                                                                        key="frame_divider",
                                                                                        default=frame_divider_default)),
                              max_lines=set_attr_resolver(instance=frame_input,
                                                          key="max_lines",
                                                          default=set_attr_resolver(instance=frame_defaults,
                                                                                    key="max_lines",
                                                                                    default=max_lines_default)),
                              multiline=set_attr_resolver(instance=frame_input,
                                                          key="multiline",
                                                          default=set_attr_resolver(instance=frame_defaults,
                                                                                    key="multiline",
                                                                                    default=multiline_default)),
                              background=set_attr_resolver(instance=frame_input,
                                                           key="background",
                                                           default=set_attr_resolver(instance=frame_defaults,
                                                                                     key="background",
                                                                                     default=background_default)),
                              trunc_value=set_attr_resolver(instance=frame_input,
                                                            key="trunc_value",
                                                            default=set_attr_resolver(instance=frame_defaults,
                                                                                      key="trunc_value",
                                                                                      default=trunc_value_default)))

    return frame_return



from tablate.type.primitives import FrameNameList
from tablate.type.type_store import FrameStoreUnion, FrameStoreList


def append_frame_to_list(frame_item: FrameStoreUnion, frame_list: FrameStoreList, frame_name_list: FrameNameList):
    pass



from tablate.classes.options.html.style.CssStyler import CssStyler

test_style = CssStyler()

test_style.add_global_style_attribute("box-sizing", "border-box")
test_style.add_global_style_attribute("margin", 0)

print(test_style.table.generate_class_names())
print(test_style.wrapper.generate_class_names())

test_style.table.add_style_attribute("background-color", "pink")
test_style.wrapper.add_style_attribute("background-color", "green")
test_style.wrapper.add_style_attribute("background-color", 0)

new_frame = test_style.frame(82).column(8).row(5)

rara = test_style.frame(2).column(4)

haha = rara.row(12).column(8)

haha.text.add_style_attribute("background-color", "blue")
haha.text.add_style_attribute("border", "none")
haha.text.add_style_attribute("text-align", "right")
haha.text.add_style_attribute("width", "100%")
haha.text.add_style_attribute("width", 0)

print(test_style.return_head_styles())

print(haha.text.generate_class_names())



from typing import Union

from tablate.classes.bases.TablateApiBase import TablateApi
from tablate.classes.helpers.get_frame import get_frame
from tablate.classes.helpers.list_frame import list_frames


class TablateSet(TablateApi):

    def list_frames(self):
        list_frames(self._frame_list)

    def get_frame(self, selector: Union[int, str]):
       return get_frame(frame_list=self._frame_list, options=self._options, selector=selector)

    def from_dict(self):
        pass

    def remove_frame(self):
        pass

    def replace_frame(self):
        pass


from typing import Union

from tablate.api.modules.Grid import Grid
from tablate.api.modules.Table import Table
from tablate.api.modules.Text import Text
from tablate.type.type_global import Options
from tablate.type.type_input import TableHeaderFrameStylesInput, HtmlTableHeaderFrameStylesInput, \
    TableBodyFrameStylesInput, HtmlTableBodyFrameStylesInput, HtmlTableFrameStylesInput, HtmlGridFrameStylesInput, \
    HtmlTextFrameStylesInput
from tablate.type.type_store import FrameStoreUnion, FrameStoreList


def get_frame(frame_list: FrameStoreList, options: Options, selector: Union[int, str]):
    selector_index = 0
    for frame_index, frame_item in enumerate(frame_list):
        if frame_item.type == "table_header":
            continue
        if (type(selector) == int and selector == selector_index) or (
                type(selector) == str and selector == frame_item.name):
            if frame_item.type == "text":
                print("here", "text")
                return Text(text=frame_item.column_list[0]["string"],
                            name=frame_item.name,
                            text_style=frame_item.text_styles.text_style,
                            text_align=frame_item.text_styles.text_align,
                            text_color=frame_item.text_styles.text_color,
                            frame_divider=frame_item.frame_styles.frame_divider,
                            background=frame_item.frame_styles.background,
                            background_padding=frame_item.column_list[0]["background_padding"],
                            multiline=frame_item.frame_styles.multiline,
                            max_lines=frame_item.frame_styles.max_lines,

                            html_px_multiplier=frame_item.html_frame_styles.html_px_multiplier,
                            html_styles=HtmlTextFrameStylesInput(html_frame_styles=frame_item.html_frame_styles,
                                                                 html_text_styles=frame_item.html_text_styles),
                            # TablateApi arge
                            outer_border=options.console.outer_styles.outer_border,
                            outer_padding=options.console.outer_styles.outer_padding,
                            outer_width=options.console.outer_styles.outer_width,
                            html_outer_styles=options.html.html_outer_styles)
            if frame_item.type == "grid":
                print("here", "grid")
                return Grid(columns=frame_item.column_list,
                            name=frame_item.name,
                            frame_divider=frame_item.frame_styles.frame_divider,
                            background=frame_item.frame_styles.background,
                            background_padding=frame_item.column_styles.background_padding,
                            multiline=frame_item.frame_styles.multiline,
                            max_lines=frame_item.frame_styles.max_lines,
                            column_divider=frame_item.column_styles.divider,
                            column_padding=frame_item.column_styles.padding,
                            text_style=frame_item.text_styles.text_style,
                            text_align=frame_item.text_styles.text_align,
                            text_color=frame_item.text_styles.text_color,
                            html_px_multiplier=frame_item.html_frame_styles.html_px_multiplier,
                            html_styles=HtmlGridFrameStylesInput(html_frame_styles=frame_item.html_frame_styles,
                                                                 html_column_styles=frame_item.html_column_styles,
                                                                 html_text_styles=frame_item.html_text_styles),
                            # TablateApi args
                            outer_border=options.console.outer_styles.outer_border,
                            outer_padding=options.console.outer_styles.outer_padding,
                            outer_width=options.console.outer_styles.outer_width,
                            html_outer_styles=options.html.html_outer_styles)
            elif frame_item.type == "table_body":
                print("here", "table")
                header_base_divider = None
                header_styles = None
                html_header_styles = None
                if frame_item.hide_header is False:
                    header_frame = frame_list[frame_index - 1]
                    header_base_divider = header_frame.frame_styles.frame_divider
                    header_styles = TableHeaderFrameStylesInput(frame_styles=header_frame.frame_styles,
                                                                column_styles=header_frame.column_styles,
                                                                text_styles=header_frame.text_styles)
                    html_header_styles = HtmlTableHeaderFrameStylesInput(
                        html_frame_styles=header_frame.html_frame_styles,
                        html_column_styles=header_frame.html_column_styles,
                        html_text_styles=header_frame.html_text_styles)
                return Table(columns=frame_item.column_list,
                             rows=frame_item.row_list,
                             name=frame_item.name,
                             frame_divider=frame_item.frame_styles.frame_divider,
                             background=frame_item.frame_styles.background,
                             background_padding=frame_item.column_styles.background_padding,
                             multiline=frame_item.frame_styles.multiline,
                             max_lines=frame_item.frame_styles.max_lines,
                             column_divider=frame_item.column_styles.divider,
                             column_padding=frame_item.column_styles.padding,
                             text_style=frame_item.text_styles.text_style,
                             text_align=frame_item.text_styles.text_align,
                             text_color=frame_item.text_styles.text_color,
                             header_base_divider=header_base_divider,
                             hide_header=frame_item.hide_header,
                             row_line_divider=frame_item.row_styles.row_line_divider,
                             odd_row_background=frame_item.row_styles.odds_background,
                             even_row_background=frame_item.row_styles.evens_background,
                             html_px_multiplier=frame_item.html_frame_styles.html_px_multiplier,
                             header_styles=header_styles,
                             body_styles=TableBodyFrameStylesInput(frame_styles=frame_item.frame_styles,
                                                                   column_styles=frame_item.column_styles,
                                                                   text_styles=frame_item.text_styles,
                                                                   row_styles=frame_item.row_styles),
                             html_header_styles=html_header_styles,
                             html_body_styles=HtmlTableBodyFrameStylesInput(
                                 html_frame_styles=frame_item.html_frame_styles,
                                 html_column_styles=frame_item.html_column_styles,
                                 html_text_styles=frame_item.html_text_styles,
                                 html_row_styles=frame_item.html_row_styles),
                             html_styles=HtmlTableFrameStylesInput(html_frame_styles=frame_item.html_frame_styles,
                                                                   html_column_styles=frame_item.html_column_styles,
                                                                   html_text_styles=frame_item.html_text_styles,
                                                                   html_row_styles=frame_item.html_row_styles),
                             # TablateApi args
                             outer_border=options.console.outer_styles.outer_border,
                             outer_padding=options.console.outer_styles.outer_padding,
                             outer_width=options.console.outer_styles.outer_width,
                             html_outer_styles=options.html.html_outer_styles)
        selector_index += 1



from typing import Union

from tablate.library.calcs.calc_column_widths import calc_column_widths
from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.library.initializers.mappers.element.style_mapper import style_mapper
from tablate.library.initializers.mappers.attribute.column_attr import column_attr
from tablate.type.defaults import column_padding_default, background_padding_default, frame_name_default
from tablate.type.type_base import FrameBase, TextBase, HtmlColumnBase, HtmlFrameBase, HtmlTextBase
from tablate.type.type_global import Options
from tablate.type.type_input import BaseStylesInput, \
    HtmlStylesInput, HtmlTextFrameStylesInput, BaseColumnInput
from tablate.type.primitives import FrameDivider, TextAlign, TextStyle, TextColor, Background, Multiline, MaxLines, \
    ColumnPadding, HtmlPxMultiplier, BackgroundPadding, FrameName
from tablate.type.type_store import GridFrameStore, FrameStore, TextStore


def text_init(text: Union[str, int, float],
              name: FrameName,
              text_style: TextStyle,
              text_align: TextAlign,
              text_color: TextColor,

              frame_divider: FrameDivider,
              frame_padding: ColumnPadding,
              background: Background,
              background_padding: BackgroundPadding,
              multiline: Multiline,
              max_lines: MaxLines,

              html_px_multiplier: HtmlPxMultiplier,
              html_styles: HtmlTextFrameStylesInput,

              options: Options) -> GridFrameStore:

    background_padding = background_padding if background_padding is not None else set_attr_resolver(instance=options.console.outer_styles,
                                                                                                     key="background_padding",
                                                                                                     default=background_padding_default)

    name = name if name is not None else frame_name_default

    html_frame_styles = html_styles.html_frame_styles if hasattr(html_styles, 'html_frame_styles') else HtmlFrameBase()
    html_column_styles = HtmlColumnBase()
    html_text_styles = html_styles.html_text_styles if hasattr(html_styles, 'html_text_styles') else HtmlTextBase()

    text_styles = style_mapper(base_input=BaseStylesInput(frame_styles=FrameStore(frame_divider=frame_divider,
                                                                                  max_lines=max_lines,
                                                                                  multiline=multiline,
                                                                                  background=background),
                                                          text_styles=TextStore(text_style=text_style,
                                                                                text_align=text_align,
                                                                                text_color=text_color)),
                               html_input=HtmlStylesInput(html_frame_styles=html_frame_styles,
                                                          html_column_styles=html_column_styles,
                                                          html_text_styles=html_text_styles),
                               frame_default=None,
                               column_default=None,
                               text_default=None,
                               html_px_multiplier=html_px_multiplier,
                               options=options)

    text_column: BaseColumnInput = {
        "string": text,
        "padding": frame_padding if frame_padding else options.console.column_styles.padding if options.console.column_styles.padding is not None else column_padding_default,
        "divider": "thin"
    }

    columns = calc_column_widths(columns=[text_column], options=options)
    # todo: pesky warning
    column = column_attr(column_dict=columns[0],
                         frame_styles=text_styles.frame_styles,
                         column_styles=text_styles.column_styles,
                         text_styles=text_styles.text_styles,
                         background_padding=background_padding)

    text_frame_store = GridFrameStore(type="text",
                                      name=name,
                                      column_list=[column],
                                      frame_styles=text_styles.frame_styles,
                                      column_styles=text_styles.column_styles,
                                      text_styles=text_styles.text_styles,
                                      html_frame_styles=text_styles.html_frame_styles,
                                      html_column_styles=text_styles.html_column_styles,
                                      html_text_styles=text_styles.html_text_styles)

    return text_frame_store



from tablate.library.renderers.console.render_console_foot import render_console_foot
from tablate.library.renderers.console.render_console_frames import render_console_frames
from tablate.library.renderers.console.render_console_head import render_console_head
from tablate.type.type_store import FrameStoreList
from tablate.type.type_global import Options


def render_console(frame_list: FrameStoreList, options: Options) -> str:
    return_string = ""
    if len(frame_list) > 0:
        return_string += render_console_head(frame_list=frame_list, options=options)
        return_string += render_console_frames(frame_list=frame_list, options=options)
        return_string += render_console_foot(frame_list=frame_list, options=options)
    return return_string



from typing import Optional

from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import multiline_default, max_lines_default, background_default, frame_divider_default, \
    html_px_multiplier_default, html_divider_weight_default
from tablate.type.primitives import HtmlPxMultiplier
from tablate.type.type_base import HtmlFrameBase, FrameBase
from tablate.type.type_store import HtmlFrameStore


def html_frame_mapper(html_frame_input: Optional[HtmlFrameBase] = None,
                      html_frame_defaults: Optional[HtmlFrameBase] = None,
                      base_frame_defaults: Optional[FrameBase] = None,
                      html_px_multiplier: HtmlPxMultiplier = html_px_multiplier_default) -> HtmlFrameStore:

    html_frame_divider_weight = set_attr_resolver(
        instance=html_frame_input,
        key="html_frame_divider_weight",
        default=set_attr_resolver(
            instance=html_frame_defaults,
            key="html_frame_divider_weight",
            default=html_divider_weight_default))
    html_frame_divider_style = set_attr_resolver(
        instance=html_frame_input,
        key="html_frame_divider_style",
        default=set_attr_resolver(
            instance=html_frame_defaults,
            key="html_frame_divider_style",
            default=set_attr_resolver(
                instance=base_frame_defaults,
                key="divider",
                default=frame_divider_default)))
    html_max_lines = set_attr_resolver(
        instance=html_frame_input,
        key="html_max_lines",
        default=set_attr_resolver(
            instance=html_frame_defaults,
            key="html_max_lines",
            default=set_attr_resolver(
                instance=base_frame_defaults,
                key="max_linnes",
                default=max_lines_default)))
    html_multiline = set_attr_resolver(
        instance=html_frame_input,
        key="html_multiline",
        default=set_attr_resolver(
            instance=html_frame_defaults,
            key="html_multiline",
            default=set_attr_resolver(
                instance=base_frame_defaults,
                key="multiline",
                default=max_lines_default)))
    html_background = set_attr_resolver(
        instance=html_frame_input,
        key="html_background",
        default=set_attr_resolver(
            instance=html_frame_defaults,
            key="html_background",
            default=set_attr_resolver(
                instance=base_frame_defaults,
                key="background",
                default=background_default)))

    html_frame_return = HtmlFrameStore(html_frame_divider_weight=html_frame_divider_weight,
                                       html_frame_divider_style=html_frame_divider_style,
                                       html_max_lines=html_max_lines,
                                       html_multiline=html_multiline,
                                       html_background=html_background,
                                       html_px_multiplier=html_px_multiplier)

    return html_frame_return



from dataclasses import dataclass, field
from typing import Optional

from tablate.type.primitives import FrameDivider, OuterBorder, TextStyle, TextAlign, ColumnWidth, \
    ColumnPadding, MaxLines, Multiline, OuterWidth, OuterPadding, HtmlTextStyle, HtmlTextColor, HtmlTextSize, \
    HtmlColumnDividerStyle, HtmlColumnPadding, HtmlBackground, HtmlFrameDivider, \
    HtmlPxMultiplier, HtmlOuterBorder, HtmlOuterPadding, HtmlOuterWidth, HtmlTextAlign, Background, TextColor, \
    TruncValue, TextString, TableRowKey, ColumnDivider, HtmlDividerWeight, BackgroundPadding


@dataclass
class OuterBase:
    outer_border: Optional[OuterBorder] = None
    outer_padding: Optional[OuterPadding] = None
    outer_width: Optional[OuterWidth] = None


###################################

@dataclass
class FrameBase:
    frame_divider: Optional[FrameDivider] = None
    max_lines: Optional[MaxLines] = None
    multiline: Optional[Multiline] = None
    background: Optional[Background] = None
    trunc_value: Optional[TruncValue] = None


@dataclass
class ColumnBase:
    divider: Optional[ColumnDivider] = None
    padding: Optional[ColumnPadding] = None
    background_padding: Optional[BackgroundPadding] = None


@dataclass
class TextBase:
    text_style: Optional[TextStyle] = None
    text_align: Optional[TextAlign] = None
    text_color: Optional[TextColor] = None


###################################
###################################
###################################

TableHeaderFrameBase = FrameBase


@dataclass
class TableRowsBase:
    row_line_divider: Optional[FrameDivider] = None
    odds_background: Optional[Background] = None
    evens_background: Optional[Background] = None


########################################################################################################################
########################################################################################################################
########################################################################################################################

@dataclass
class HtmlOuterBase:
    html_outer_border_weight: Optional[HtmlDividerWeight] = None
    html_outer_border_style: Optional[HtmlOuterBorder] = None
    html_outer_padding: Optional[HtmlOuterPadding] = None
    html_outer_width: Optional[HtmlOuterWidth] = None

@dataclass
class HtmlFrameBase:
    html_frame_divider_style: Optional[HtmlFrameDivider] = None
    html_frame_divider_weight: Optional[HtmlDividerWeight] = None
    html_max_lines: Optional[MaxLines] = None
    html_multiline: Optional[Multiline] = None
    html_background: Optional[HtmlBackground] = None


@dataclass
class HtmlColumnBase:
    html_divider_style: Optional[HtmlColumnDividerStyle] = None
    html_divider_weight: Optional[HtmlDividerWeight] = None
    html_padding: Optional[HtmlColumnPadding] = None


@dataclass
class HtmlTextBase:
    html_text_style: Optional[HtmlTextStyle] = None
    html_text_align: Optional[HtmlTextAlign] = None
    html_text_color: Optional[HtmlTextColor] = None
    html_text_size: Optional[HtmlTextSize] = None


###################################
###################################
###################################


@dataclass
class HtmlTableRowsBase:
    html_row_line_divider_weight: Optional[HtmlDividerWeight] = None
    html_row_line_divider_style: Optional[HtmlFrameDivider] = None
    html_odds_background: Optional[HtmlBackground] = None
    html_evens_background: Optional[HtmlBackground] = None






from dataclasses import dataclass
from typing import List, Optional

from tablate.classes.options.html.style.CssStyler import CssStyler
from tablate.type.primitives import CssStyleBlock, PxInteger
from tablate.type.type_base import HtmlOuterBase, FrameBase, ColumnBase, TextBase, OuterBase, HtmlFrameBase, \
    HtmlColumnBase, HtmlTextBase
from tablate.type.type_store import HtmlOuterStore, OuterStore


@dataclass
class ConsoleOptions:
    outer_styles: OuterStore
    frame_styles: FrameBase
    column_styles: ColumnBase
    text_styles: TextBase


@dataclass
class HtmlOptions:
    html_outer_styles: HtmlOuterStore
    html_frame_styles: HtmlFrameBase
    html_column_styles: HtmlColumnBase
    html_text_styles: HtmlTextBase
    css_injection: CssStyleBlock
    styler: Optional[CssStyler]
    column_baselines: List[PxInteger]


@dataclass
class Options:
    console: ConsoleOptions
    html: HtmlOptions




from tablate.library.ascii.chars.line_v import v_line
from tablate.library.checkers.is_last_element import is_last_element
from tablate.library.formatters.console.cell_string import cell_string_single_line, cell_string_multi_line
from tablate.library.renderers.console.frames.render_console_columns import column_console_multiline
from tablate.library.formatters.console.row_outer_border import row_outer_border
from tablate.type.type_store import GridFrameStore
from tablate.type.type_global import Options


def render_console_single_line_grid(grid_frame_dict: GridFrameStore, options: Options):

    grid_line_string = ""
    for column_index, column_item in enumerate(grid_frame_dict.column_list):
        grid_line_string += cell_string_single_line(string=column_item["string"],
                                                    column_item=column_item,
                                                    trunc_value=grid_frame_dict.frame_styles.trunc_value)
        if not is_last_element(column_index, grid_frame_dict.column_list):
            grid_line_string += v_line[column_item["divider"]]

    return row_outer_border(row_string=grid_line_string, options=options)


def render_console_multi_line_grid(grid_frame_dict: GridFrameStore, options: Options):

    formatted_columns_array = []

    for column_item in grid_frame_dict.column_list:
        column_string_array = cell_string_multi_line(string=column_item["string"],
                                                     column_item=column_item,
                                                     trunc_value=grid_frame_dict.frame_styles.trunc_value,
                                                     max_lines=grid_frame_dict.frame_styles.max_lines)
        formatted_columns_array.append(column_string_array)

    return column_console_multiline(formatted_columns_array=formatted_columns_array,
                                    frame_dict=grid_frame_dict,
                                    options=options)



from __future__ import annotations

column_list1 = [
    {
        "key": "A really really long key like this - maybe even longer",
        "width":  30
    },
    {
        "key": "Key Two"
    },
    {
        "key": "Key Three",
        "width": 20
    }
]

row_list1 = [
    {
        "A really really long key like this - maybe even longer": "Quam elementum pulvinar etiam non quam lacus suspendisse faucibus. Enim nec dui nunc mattis enim ut tellus. Est placerat in egestas erat imperdiet sed euismod nisi porta. At tempor commodo ullamcorper a lacus vestibulum sed. Porta nibh venenatis cras sed. Cras sed felis eget velit aliquet sagittis id consectetur. Aliquam id diam maecenas ultricies mi. ",
        "Key Two": "Quam elementum pulvinar etiam non quam lacus suspendisse faucibus. Enim nec dui nunc mattis enim ut tellus. Est placerat in egestas erat imperdiet sed euismod nisi porta. At tempor commodo ullamcorper a lacus vestibulum sed. Porta nibh venenatis cras sed. Cras sed felis eget velit aliquet sagittis id consectetur. Aliquam id diam maecenas ultricies mi. ",
        "Key Three": "Quam elementum pulvinar etiam non quam lacus suspendisse faucibus. Enim nec dui nunc mattis enim ut tellus. Est placerat in egestas erat imperdiet sed euismod nisi porta. At tempor commodo ullamcorper a lacus vestibulum sed. Porta nibh venenatis cras sed. Cras sed felis eget velit aliquet sagittis id consectetur. Aliquam id diam maecenas ultricies mi. ",

    },
    {
        "A really really long key like this - maybe even longer": "Row Two Column Two",
        "Key Two": "Row Two Column Two",
        "Key Three": "Row Two Column Three"

    },
    {
        "A really really long key like this - maybe even longer": "Row Three Column One",
        "Key Two": "Row Three Column Two",
        "Key Three": "Row Three Column Three"

    },
]

column_list2a = [
    {
        "key": "Key One",
        "width": "50%",
        "divider": "thick"
    },
    {
        "key": "Key Two",
        "text_align": "right"
    },
    {
        "key": "Key Three",
        "text_align": "right"
    }
]

column_list2b = [
    {
        "key": "Key One",
        "width": "20%",

    },
    {
        "key": "Key Two",
        "text_align": "right"
    },
    {
        "key": "Key Three",
        "text_align": "right"
    }
]

column_list2c = [
    {
        "key": "Key One",
        "width": "30%",
        "divider": "thick"
    },
    {
        "key": "Key Two",
        "text_align": "right",
        "divider": "thin"
    },
    {
        "key": "Key Three",
        "text_align": "right"
    }
]

row_list2 = [
    {
        "Key One": "Some string value",
        "Key Two": 4,
        "Key Three": 9.6

    },
    {
        "Key One": "Some other string value",
        "Key Two": 5,
        "Key Three": 4.6

    },
    {
        "Key One": "Some final string value",
        "Key Two": 100,
        "Key Three": 328.832

    }
]

column_list3 = [
    {
        "string": "Quam elementum pulvinar etiam non quam lacus suspendisse faucibus. Enim nec dui nunc mattis enim ut tellus. Est placerat in egestas erat imperdiet sed euismod nisi porta. At tempor commodo ullamcorper a lacus vestibulum sed. Porta nibh venenatis cras sed. Cras sed felis eget velit aliquet sagittis id consectetur. Aliquam id diam maecenas ultricies mi. ",
        "width": 40,
        "divider": "thick"
    },
    {
        "string": "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.",
    },
    {
        "string": "Commodo elit at imperdiet dui accumsan sit amet nulla. Sodales neque sodales ut etiam sit amet nisl purus. Convallis posuere morbi leo urna molestie at elementum. Enim nunc faucibus a pellentesque sit amet porttitor. Enim nulla aliquet porttitor lacus luctus accumsan tortor. Ullamcorper velit sed ullamcorper morbi. Sit amet porttitor eget dolor morbi non arcu. Eget mauris pharetra et ultrices neque ornare aenean euismod elementum. Euismod lacinia at quis risus sed vulputate odio. Varius quam quisque id diam. Habitant morbi tristique senectus et netus et malesuada fames ac. Ornare quam viverra orci sagittis eu volutpat odio. Nisl suscipit adipiscing bibendum est ultricies. Adipiscing elit pellentesque habitant morbi tristique senectus. Tellus cras adipiscing enim eu turpis egestas pretium aenean pharetra. At consectetur lorem donec massa sapien faucibus et molestie. Neque ornare aenean euismod elementum nisi quis eleifend quam. A arcu cursus vitae congue mauris."
    }
]

column_list4 = [
    {
        "string": "Quam elementum pulvinar etiam non quam lacus suspendisse faucibus. Enim nec dui nunc mattis enim ut tellus. Est placerat in egestas erat imperdiet sed euismod nisi porta. At tempor commodo ullamcorper a lacus vestibulum sed. Porta nibh venenatis cras sed. Cras sed felis eget velit aliquet sagittis id consectetur. Aliquam id diam maecenas ultricies mi. ",
        "width": 20
    },
    {
        "string": "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.",
        "divider": "thick"
    },
    {
        "string": "Commodo elit at imperdiet dui accumsan sit amet nulla. Sodales neque sodales ut etiam sit amet nisl purus. Convallis posuere morbi leo urna molestie at elementum. Enim nunc faucibus a pellentesque sit amet porttitor. Enim nulla aliquet porttitor lacus luctus accumsan tortor. Ullamcorper velit sed ullamcorper morbi. Sit amet porttitor eget dolor morbi non arcu. Eget mauris pharetra et ultrices neque ornare aenean euismod elementum. Euismod lacinia at quis risus sed vulputate odio. Varius quam quisque id diam. Habitant morbi tristique senectus et netus et malesuada fames ac. Ornare quam viverra orci sagittis eu volutpat odio. Nisl suscipit adipiscing bibendum est ultricies. Adipiscing elit pellentesque habitant morbi tristique senectus. Tellus cras adipiscing enim eu turpis egestas pretium aenean pharetra. At consectetur lorem donec massa sapien faucibus et molestie. Neque ornare aenean euismod elementum nisi quis eleifend quam. A arcu cursus vitae congue mauris."
    }
]

column_list5 = [
    {
        "string": 123,
        "width": "20%",
        "divider": "thick"
    },
    {
        "string": 456,

    },
    {
        "string": 789,
        "width": "25%"
    }
]

really_long_string = "Commodo elit at imperdiet dui accumsan sit amet nulla. Sodales neque sodales ut etiam sit amet nisl purus. Convallis posuere morbi leo urna molestie at elementum. Enim nunc faucibus a pellentesque sit amet porttitor. Enim nulla aliquet porttitor lacus luctus accumsan tortor. Ullamcorper velit sed ullamcorper morbi. Sit amet porttitor eget dolor morbi non arcu. Eget mauris pharetra et ultrices neque ornare aenean euismod elementum. Euismod lacinia at quis risus sed vulputate odio. Varius quam quisque id diam. Habitant morbi tristique senectus et netus et malesuada fames ac. Ornare quam viverra orci sagittis eu volutpat odio. Nisl suscipit adipiscing bibendum est ultricies. Adipiscing elit pellentesque habitant morbi tristique senectus. Tellus cras adipiscing enim eu turpis egestas pretium aenean pharetra. At consectetur lorem donec massa sapien faucibus et molestie. Neque ornare aenean euismod elementum nisi quis eleifend quam. A arcu cursus vitae congue mauris."



from typing import Optional

from tablate.library.checkers.set_attr_resolver import set_attr_resolver
from tablate.type.defaults import column_divider_default, background_default, column_padding_default
from tablate.type.type_base import ColumnBase
from tablate.type.type_store import ColumnStore


def base_column_mapper(columns_input: Optional[ColumnBase] = None,
                       column_defaults: Optional[ColumnBase] = None) -> ColumnStore:
    if columns_input is None:
        columns_input = ColumnBase()

    columns_return = ColumnStore(divider=set_attr_resolver(instance=columns_input,
                                                           key="divider",
                                                           default=set_attr_resolver(instance=column_defaults,
                                                                                     key="divider",
                                                                                     default=column_divider_default)),
                                 padding=set_attr_resolver(instance=columns_input,
                                                           key="padding",
                                                           default=set_attr_resolver(instance=column_defaults,
                                                                                     key="padding",
                                                                                     default=column_padding_default)))

    return columns_return



from tablate.classes.options.html.style.CssStyler import CssStyler
from tablate.library.calcs.calc_column_percent import calc_column_percent
from tablate.library.renderers.html.render_html_foot import render_html_foot
from tablate.library.renderers.html.render_html_frames import render_html_frames
from tablate.library.renderers.html.render_html_head import render_html_head
from tablate.type.type_store import FrameStoreList
from tablate.type.type_global import Options


def render_html(frame_list: FrameStoreList, options: Options) -> str:

    options.html.styler = CssStyler()

    options.html.styler.inject_css_block(options.html.css_injection)

    return_html = ""

    if len(frame_list) > 0:
        frame_list, column_baselines = calc_column_percent(frame_list=frame_list,
                                                           outer_width=options.console.outer_styles.outer_width)
        options.html.column_baselines = column_baselines
        return_html += render_html_head(options=options)
        return_html += render_html_frames(frame_list=frame_list, options=options)
        return_html += render_html_foot()

    css_head = options.html.styler.return_head_styles()
    css_foot = options.html.styler.return_foot_styles()

    return_html = f"{css_head}{return_html}{css_foot}"

    return return_html



from typing import List, Dict, Union

from tablate.classes.bases.TablateApiBase import TablateApi
from tablate.library.initializers.table_init import table_init
from tablate.type.primitives import FrameDivider, ColumnDivider, HeaderAlignment, HeaderColumnDivider, OuterBorder, \
    FrameName, Multiline, MaxLines, Background, BackgroundPadding, HideHeader, ColumnPadding, TextAlign, TextStyle, \
    TextColor, HtmlPxMultiplier, OuterPadding, OuterWidth
from tablate.type.type_input import TableColumnInput, TableRowsDataInputDict, TableBodyFrameStylesInput, \
    TableHeaderFrameStylesInput, HtmlTableFrameStylesInput, HtmlTableHeaderFrameStylesInput, \
    HtmlTableBodyFrameStylesInput, HtmlOuterStylesInput


class Table(TablateApi):

    def __init__(self,
                 # TablateTable args
                 columns: List[TableColumnInput],
                 rows: List[TableRowsDataInputDict],

                 name: FrameName = None,

                 frame_divider: FrameDivider = None,
                 multiline: Multiline = None,
                 max_lines: MaxLines = None,
                 background: Background = None,
                 background_padding: BackgroundPadding = None,

                 multiline_header: Multiline = None,
                 max_lines_header: MaxLines = None,
                 hide_header: HideHeader = None,

                 column_divider: ColumnDivider = None,
                 column_padding: ColumnPadding = None,
                 header_base_divider: FrameDivider = None,

                 row_line_divider: FrameDivider = None,
                 odd_row_background: Background = None,
                 even_row_background: Background = None,

                 text_style: TextStyle = None,
                 text_align: TextAlign = None,
                 text_color: TextColor = None,

                 header_styles: TableHeaderFrameStylesInput = None,
                 body_styles: TableBodyFrameStylesInput = None,

                 html_px_multiplier: HtmlPxMultiplier = None,
                 html_styles: HtmlTableFrameStylesInput = None,

                 html_header_styles: HtmlTableHeaderFrameStylesInput = None,
                 html_body_styles: HtmlTableBodyFrameStylesInput = None,

                 outer_border: OuterBorder = None,
                 outer_padding: OuterPadding = None,
                 outer_width: OuterWidth = None,
                 html_outer_styles: HtmlOuterStylesInput = None) -> None:
        TablateApi.__init__(self=self,
                            outer_border=outer_border,
                            outer_padding=outer_padding,
                            frame_divider=frame_divider,
                            outer_width=outer_width,
                            html_styles=html_outer_styles)

        table_head_dict, table_rows_dict = table_init(columns=columns,
                                                      rows=rows,
                                                      name=name,
                                                      frame_divider=frame_divider,
                                                      multiline=multiline,
                                                      max_lines=max_lines,
                                                      background=background,
                                                      background_padding=background_padding,
                                                      multiline_header=multiline_header,
                                                      max_lines_header=max_lines_header,
                                                      hide_header=hide_header,
                                                      column_divider=column_divider,
                                                      column_padding=column_padding,
                                                      header_base_divider=header_base_divider,
                                                      row_line_divider=row_line_divider,
                                                      odd_row_background=odd_row_background,
                                                      even_row_background=even_row_background,
                                                      text_style=text_style,
                                                      text_align=text_align,
                                                      text_color=text_color,
                                                      header_styles=header_styles,
                                                      body_styles=body_styles,
                                                      html_px_multiplier=html_px_multiplier,
                                                      html_styles=html_styles,
                                                      html_header_styles=html_header_styles,
                                                      html_body_styles=html_body_styles,
                                                      options=self._options)
        if table_head_dict:
            self._frame_list.append(table_head_dict)

        self._frame_list.append(table_rows_dict)



from typing import List

from tablate.classes.bases.TablateApiSet import TablateSet
from tablate.type.primitives import FrameDivider, OuterBorder


def concat(tablate_set: List[TablateUnion],
           outer_border: OuterBorder = "thick",
           outer_padding: int = 1,
           frame_divider: FrameDivider = "thick",
           outer_width: int = None,
           html_px_size: int = 6,
           html_text_size: int = 12,
           html_outer_width: str = "100%",
           html_css_injection: str = ""):

    return TablateSet(tablate_set=tablate_set,
                      outer_border=outer_border,
                      outer_padding=outer_padding,
                      frame_divider=frame_divider,
                      outer_width=outer_width,
                      html_px_size=html_px_size,
                      html_text_size=html_text_size,
                      html_outer_width=html_outer_width,
                      html_css_injection=html_css_injection)


from typing import Union


def concat_string(string: Union[str, int], width: int, padding: int, trunc_value: str):
    string = str(string)
    if len(string) + (padding * 2) > width:
        slice_outer_index = width - ((padding * 2) + len(trunc_value))
        if slice_outer_index < 1:
            trunc_value = trunc_value[0:len(trunc_value) - abs(slice_outer_index - 1)]
            slice_outer_index = 1
        string = f"{string[0:slice_outer_index]}{trunc_value}"

    white_space = width - ((padding * 2) + len(string))

    return string, white_space